@@ -14,3 +14,7 @@ class FlipSideApiError(Exception):
     def __init__(self, e):
         message = f"An api error occurred: {str(e)}"
         super().__init__(message)
+
+
+class PostgresConstraintError(Exception):
+    pass
