@@ -1,12 +1,13 @@
 from datetime import timedelta
-from typing import Dict, Callable, List, Tuple, Any
+from typing import Any, Callable, Dict, List, Literal, Tuple
 
 import pandas as pd
 from pandas import DataFrame
-from pyspark.sql import DataFrame as PySparkDataFrame, Window
+from pyspark.sql import DataFrame as PySparkDataFrame
+from pyspark.sql import Window
 from pyspark.sql import functions as F
-from pyspark.sql.functions import array_distinct, array_union, coalesce, array, lit
-from pyspark.sql.types import IntegerType, StructType, StructField
+from pyspark.sql.functions import array, array_distinct, array_union, coalesce, lit
+from pyspark.sql.types import IntegerType, StructField, StructType
 
 from seshat.data_class import SFrame, SPFrame
 from seshat.data_class.base import GroupSFrame
@@ -15,9 +16,9 @@ from seshat.general import configs
 from seshat.general.exceptions import InvalidArgumentsError
 from seshat.transformer import Transformer
 from seshat.transformer.merger import Merger
-from seshat.transformer.schema import Schema, Col
-from seshat.utils import pandas_func
-from seshat.utils import pyspark_func
+from seshat.transformer.schema import Col, Schema
+from seshat.transformer.trimmer.base import FilterTrimmer
+from seshat.utils import pandas_func, pyspark_func
 from seshat.utils.validation import NumericColumnValidator, TimeStampColumnValidator
 
 SYMBOLS_RECEIVED_COL = "symbols_received"
@@ -1091,54 +1092,32 @@ class GroupByDeriverMeanMax(SFrameDeriver):
 
 class TimeWindowTransformer(SFrameDeriver):
     """
-    Applies transformations to data within specified time windows and merges the results.
+    A class that derives time series-based features from a default SFrame using a transformer and merger strategy.
+    It supports applying transformations over multiple durations and combines the results using a merger.
 
-    This transformer filters data based on time windows defined by durations, applies a
-    specified transformer to each filtered dataset, and merges the results using a provided merger.
-    It's particularly useful for generating time-based features or analyzing data across
-    different time periods.
+    Duration labeling can be handled in two ways:
+    1. By adding a new column with the name specified in `label_col`, where each row is labeled with its corresponding duration.
+    2. By appending the duration label as a suffix to the names of the new columns created by the transformer, excluding those required for merging.
+
+    This behavior is controlled by the `specify_duration_by` parameter.
 
     Parameters
     ----------
-    transformer : Transformer
-        The transformer to apply to each time window's data.
+    transformer : Transformer or Callable
+        An instance of a transformer or a callable function that performs the transformation on the input data.
     merger : Merger
-        The merger used to combine results from different time windows.
-        Will be set to work in-place.
+        An instance of a merger used to combine features generated across different durations.
     time_col : str, optional
-        The column name containing timestamp data.
-        Default is configs.BLOCK_TIMESTAMP_COL.
-    durations : List[Tuple[str, int]], optional
-        List of tuples where each tuple contains a label and a duration in minutes.
-        The label identifies the time window and the duration specifies how far back
-        from the latest timestamp to include data.
+        The name of the timestamp column in the SFrame. Defaults to `configs.BLOCK_TIMESTAMP_COL`.
+    durations : List[Tuple[str, int]]
+        A list of (label, duration) pairs where each label identifies a time window (in minutes) for feature extraction.
     label_col : str, optional
-        The column name to store the time window label.
-        Default is configs.DURATION_LABEL_COL.
-    group_keys : dict, optional
-        Keys used to identify and retrieve data from a grouped SFrame.
-
-    Attributes
-    ----------
-    time_col : str
-        The column name containing timestamp data.
-    transformer : Transformer
-        The transformer applied to each time window's data.
-    label_col : str
-        The column name storing the time window label.
-    merger : Merger
-        The merger used to combine results from different time windows.
-    periods : List[Tuple[str, int]]
-        The list of time windows with their labels and durations.
-
-    Notes
-    -----
-    The transformer processes data by:
-    1. Finding the latest timestamp in the data
-    2. Creating time windows based on the specified durations
-    3. Filtering data for each time window
-    4. Applying the transformer to each filtered dataset
-    5. Merging the results using the provided merger
+        The name of the column that will store the duration label for each record. Defaults to `configs.DURATION_LABEL_COL`.
+    specify_duration_by : {'column', 'suffix'}, default 'column'
+        Specifies how duration information is represented in the output:
+        * 'column' – Adds a new column (named by `label_col`) containing the duration label.
+        * 'suffix' – Appends the duration label as a suffix to the names of columns created by the transformer,
+          excluding those required by the merger.
     """
 
     DEFAULT_GROUP_KEYS = {
@@ -1148,11 +1127,12 @@ class TimeWindowTransformer(SFrameDeriver):
 
     def __init__(
         self,
-        transformer: Transformer,
+        transformer: Transformer | Callable,
         merger: Merger,
+        durations: List[Tuple[str, int]],
         time_col=configs.BLOCK_TIMESTAMP_COL,
-        durations: List[Tuple[str, int]] = None,
         label_col=configs.DURATION_LABEL_COL,
+        specify_duration_by: Literal["column", "suffix"] = "column",
         group_keys=None,
     ):
         super().__init__(group_keys)
@@ -1160,191 +1140,89 @@ class TimeWindowTransformer(SFrameDeriver):
         self.transformer = transformer
         self.label_col = label_col
         self.merger = merger
-        self.periods = durations
+        self.specify_duration_by = specify_duration_by
+        self.durations = durations
 
-        # Ensure that merger work inplace
+        # Ensure that the merge works in place and group keys are set to their default values.
         self.merger.inplace = True
+        self.merger.group_keys = {"default": "default", "other": "other"}
+        if self.specify_duration_by == "column":
+            self.merger.axis = 0
 
     def validate(self, sf: SFrame):
-        """
-        Validate that the input SFrame contains the required columns.
-
-        Parameters
-        ----------
-        sf : SFrame
-            The input SFrame to validate.
-
-        Raises
-        ------
-        ColDoesNotExistError
-            If the required time column is not present in the SFrame.
-        """
         super().validate(sf)
         self._validate_columns(sf, self.group_keys["default"], self.time_col)
 
-    def calculate_metrics_df(self, default: DataFrame, label, start) -> SFrame:
-        """
-        Calculate metrics for a specific time window.
+    def _is_empty(self, sf: SFrame) -> bool:
+        raw = sf.to_raw()
+        if raw is None:
+            return True
+        elif hasattr(raw, "rdd"):
+            return raw.rdd.isEmpty()
+        else:
+            return len(raw) == 0
 
-        Parameters
-        ----------
-        default : DataFrame
-            The input DataFrame containing the data.
-        label : str
-            The label for the time window.
-        start : datetime
-            The start timestamp for the time window.
+    def calculate_metrics(self, sf: SFrame, label, start) -> SFrame:
+        trimmer = FilterTrimmer(op=">=", col=self.time_col, value=start)
+        filtered = trimmer(sf)
 
-        Returns
-        -------
-        SFrame
-            An SFrame containing the calculated metrics for the time window.
-            Returns an empty DFrame if no data is available for the time window.
-        """
-        filtered = default[default[self.time_col] >= start]
+        if self._is_empty(filtered):
+            return filtered
 
-        if len(filtered) == 0:
-            return DFrame(DataFrame())
+        result = self.transformer(sf_input=filtered, label=label, start=start)
+        if self.specify_duration_by == "column":
+            deriver = StaticValueColumnAdder(col_name=self.label_col, value=label)
+            result = deriver(result)
 
-        filtered.loc[:, self.label_col] = label
-        return self.transformer(DFrame(filtered), label=label, start=start)
+        else:
+            # Find columns excluding those used by,
+            # the merger for merging.
+            cols = set(result[self.transformer].get_columns()) - set(self.merger.on)
+            schema = Schema(
+                cols=[Col(col, to=f"{col}_{label}") for col in cols],
+                exclusive=False,
+            )
+            result = schema(result)
+        return result
 
     def derive_df(self, default: DataFrame, *args, **kwargs):
-        """
-        Derive time-based features from the input DataFrame.
-
-        This method:
-        1. Validates that the time column contains timestamp data
-        2. Finds the latest timestamp in the data
-        3. Creates time windows based on the specified durations
-        4. Calculates metrics for each time window
-        5. Merges the results using the provided merger
-
-        Parameters
-        ----------
-        default : DataFrame
-            The input DataFrame containing the data.
-        *args : tuple
-            Variable length argument list.
-        **kwargs : dict
-            Arbitrary keyword arguments.
-
-        Returns
-        -------
-        dict
-            A dictionary with two keys:
-            - 'default': The original input DataFrame
-            - 'timeseries_feature': The derived time-based features
-        """
         default = TimeStampColumnValidator().validate(default, self.time_col)
-
         last_timestamp = default[self.time_col].max()
+        sf_input = DFrame.from_raw(default)
+        result = DFrame()
+        return self.derive(sf_input, result, last_timestamp)
+
+    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
+        default = TimeStampColumnValidator().validate(default, self.time_col)
+        last_timestamp = default.select(F.max(self.time_col)).collect()[0][0]
+        sf_input = SPFrame.from_raw(default)
+        result = SPFrame()
+        return self.derive(sf_input, result, last_timestamp)
+
+    def derive(self, sf_input: SFrame, result: SFrame, last_timestamp):
         intervals = {
             label: last_timestamp - timedelta(minutes=minutes)
-            for label, minutes in self.periods
+            for label, minutes in self.durations
         }
 
-        result = DFrame(DataFrame())
-        for label, start in intervals.items():
-            metrics = self.calculate_metrics_df(default, label, start)
-            if metrics.data.empty:
+        for idx, (label, start) in enumerate(intervals.items()):
+            metrics = self.calculate_metrics(sf_input, label, start)
+            if self._is_empty(metrics):
                 continue
             metrics = metrics.make_group()
-            metrics[self.merger.group_keys["other"]] = result
+            metrics["other"] = metrics["default"]
+            metrics["default"] = result
             result = (
-                metrics["default"]
-                if result.data.empty
+                metrics["other"]
+                if self._is_empty(result)
                 else self.merger(metrics)["default"]
             )
 
-        return {"default": default, "timeseries_feature": result.to_raw()}
-
-    def calculate_metrics_spf(
-        self, default: PySparkDataFrame, label, start
-    ) -> PySparkDataFrame:
-        """
-        Calculate metrics for a specific time window using PySpark DataFrame.
-
-        Parameters
-        ----------
-        default : PySparkDataFrame
-            The input PySpark DataFrame containing the data.
-        label : str
-            The label for the time window.
-        start : datetime
-            The start timestamp for the time window.
-
-        Returns
-        -------
-        PySparkDataFrame
-            A PySpark DataFrame containing the calculated metrics for the time window.
-            Returns an empty DataFrame if no data is available for the time window.
-        """
-        filtered = default.filter(F.col(self.time_col) >= start)
-
-        if filtered.count() == 0:
-            return SPFrame.get_spark().createDataFrame([], default.schema)
-
-        filtered = filtered.withColumn(self.label_col, F.lit(label))
-        return self.transformer(SPFrame(filtered), label=label, start=start).data
-
-    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
-        """
-        Derive time-based features from the input PySpark DataFrame.
-
-        This method:
-        1. Validates that the time column contains timestamp data
-        2. Finds the latest timestamp in the data
-        3. Creates time windows based on the specified durations
-        4. Calculates metrics for each time window
-        5. Merges the results using the provided merger
-
-        Parameters
-        ----------
-        default : PySparkDataFrame
-            The input PySpark DataFrame containing the data.
-        *args : tuple
-            Variable length argument list.
-        **kwargs : dict
-            Arbitrary keyword arguments.
-
-        Returns
-        -------
-        dict
-            A dictionary with two keys:
-            - 'default': The original input DataFrame
-            - 'timeseries_feature': The derived time-based features
-        """
-        default = TimeStampColumnValidator().validate(default, self.time_col)
-
-        last_timestamp_row = default.agg(F.max(self.time_col)).collect()[0][0]
-
-        intervals = {
-            label: last_timestamp_row - timedelta(minutes=minutes)
-            for label, minutes in self.periods
+        return {
+            "default": sf_input[self.default_sf_key].to_raw(),
+            "timeseries_feature": result.to_raw(),
         }
 
-        result = None
-
-        for label, start in intervals.items():
-            metrics = self.calculate_metrics_spf(default, label, start)
-
-            if metrics.count() == 0:
-                continue
-
-            metrics_group = {self.merger.group_keys["default"]: metrics}
-
-            if result is not None:
-                metrics_group[self.merger.group_keys["other"]] = result
-                result = self.merger(SPFrame(metrics_group))["default"]
-            else:
-                result = metrics
-
-        if result is not None:
-            return {"default": default, "timeseries_feature": result}
-        else:
-            return {"default": default}
-
 
 class StaticValueColumnAdder(SFrameDeriver):
     """
@@ -1714,9 +1592,11 @@ class ComprehensiveFeaturesDeriver(SFrameDeriver):
                     last_price=("TOKEN_PRICE", "last"),
                     transaction_count_hourly=(
                         "BLOCK_TIMESTAMP",
-                        lambda x: len(x) / ((x.max() - x.min()).total_seconds() / 3600)
-                        if (x.max() > x.min())
-                        else len(x),
+                        lambda x: (
+                            len(x) / ((x.max() - x.min()).total_seconds() / 3600)
+                            if (x.max() > x.min())
+                            else len(x)
+                        ),
                     ),
                 )
                 .reset_index()
